import os
from os.path import isfile, isdir

def create_parentdir(path):
    """Create parent directory for a file/directory if not exists
    
    Args:
        path ([str]): a file path
    """
    dir = os.path.dirname(path)
    if not os.path.isdir(dir):
        os.makedirs(dir)

def pretty_table(title, content):
    """Print a pretty table according to the title and content
    
    Args:
        title ([str]): a list containing the title of table, e.g. ['title1', 'title2']
        content ([str]): a nested list containing the content of table, e.g. [['a1', 'a2'], ['b1', 'b2']]
    """
    title_ncol = len(title)
    content_ncols = map(len, content)
    if not all([title_ncol == i for i in content_ncols]):
        raise ValueError("Input content must have the same number of columns as title")
    content.insert(0, title)

    table_t = list(map(list, zip(*content)))
    col_width = list(map(lambda x: max([len(str(i)) for i in x]), table_t))
    
    def print_mark_row(col_width):
        for i in col_width:
            print("+"+"-"*i, end="")
        print("+")
    
    print_mark_row(col_width)
    for row in content:
        for i, text in enumerate(row):
            print("|"+str(text)+" "*(col_width[i] - len(str(text))), end="")
        print("|")
        print_mark_row(col_width)

    return

if __name__ == "__main__":
    # Test pretty_table
    title = ['Alias', 'Username', 'IP address', 'Port']
    content = [['host1', 'wxx', '127.0.0.1', '22'],
                ['h2', 'asdfdsg', '134.23.41.34', '22']]
    pretty_table(title, content)




# -*- coding: utf-8 -*-

import pytest
from loon.skeleton import fib

__author__ = "ShixiangWang"
__copyright__ = "ShixiangWang"
__license__ = "mit"


# def test_fib():
#     assert fib(1) == 1
#     assert fib(2) == 1
#     assert fib(7) == 13
#     with pytest.raises(AssertionError):
#         fib(-10)



# -*- coding: utf-8 -*-
"""
The skeleton of loon package.
Set command in [options.entry_points] section of setup.cfg file
Install with `python setup.py install`
"""

import os
import sys
import argparse
import logging

if __package__ == '' or __package__ is None:  # Use for test
  from __init__ import __version__, __author__, __license__
  from classes import Host
else:
  from loon import __version__, __author__, __license__
  from loon.classes import Host

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Be an efficient loon.")

    # Show version info
    parser.add_argument(
        "--version",
        action="version",
        version="loon {ver} released under {license} license.".format(ver=__version__, license=__license__.upper()))
    parser.add_argument(
        "--author",
        action="version",
        help="show info of program's author",
        version="Author: 王诗翔 Email: w_shixiang@163.com GitHub: @{author}".format(author=__author__))

    # Common arguments
    host_parent_parser = argparse.ArgumentParser(add_help=False)
    host_parent_parser.add_argument(
      '-N', '--name',
      dest="name",
      help='Host alias, default is value from -U',
      type = str
    )
    host_parent_parser.add_argument(
      '-U', '--username',
      dest='username',
      help='Username for remote host',
      type = str)
    host_parent_parser.add_argument(
      '-H', '--host',
      dest='host',
      help='IP address for remote host (e.g. 192.168.0.1)',
      type = str)
    host_parent_parser.add_argument(
      '-P', '--port',
      dest="port",
      help='Port for remote host, default is 22', 
      type = int,
      default=22)


    verbose_parser = argparse.ArgumentParser(add_help=False)
    verbose_parser.add_argument(
      "-v",
      "--verbose",
      dest="loglevel",
      help="set loglevel to INFO",
      action="store_const",
      const=logging.INFO)

    # Subcommands
    subparsers = parser.add_subparsers(
      title='subcommands',
      #description='valid subcommands',
      help="description",
      dest="subparsers_name")

    # Create the parser for the "add" command
    parser_add = subparsers.add_parser(
      'add',
      help="Add a remote host",
      parents=[host_parent_parser, verbose_parser])
    parser_add.add_argument(
      '-A',
      '--active',
      dest='switch_active',
      help='Set new host as active host',
      action='store_true')
    
    # Create the parser for the "add" command
    parser_del = subparsers.add_parser(
      'delete',
      help="Delete a remote host",
      parents=[host_parent_parser, verbose_parser])

    # Create the parser for the "switch" command
    parser_switch = subparsers.add_parser(
      'switch',
      help="Switch active remote host",
      parents=[host_parent_parser, verbose_parser])

    # Create the parser for the "list" command
    parser_list = subparsers.add_parser(
      'list',
      help="List all remote hosts",
      parents=[verbose_parser])
    
    # Create the parser for the "rename" command
    parser_rename = subparsers.add_parser(
      'rename',
      help="Rename host alias",
      parents=[verbose_parser])
    parser_rename.add_argument(
      'old',
      help="Old host alias",
      type=str
    )
    parser_rename.add_argument(
      'new',
      help="New host alias",
      type=str
    )

    # Create the parser for the "run" command
    parser_run = subparsers.add_parser(
      'run',
      help='Run commands on the active remote host',
      parents=[verbose_parser]
    )
    parser_run.add_argument(
      nargs='+',
      dest='commands',
      help="Commands to run, special symbol or option should be quoted, e.g. 'ls ~', 'ls -l'",
      type=str
    )
    
    # Create the parser for the "upload" command
    parser_upload = subparsers.add_parser(
      'upload',
      help='Upload files to active remote host',
      parents=[verbose_parser]
    )
    parser_upload.add_argument(
      'source',
      nargs='+',
      help='Source files to upload'
    )
    parser_upload.add_argument(
      'destination',
      help="Remote destination directory, Note'~' should be quoted",
      type=str
    )

    return parser.parse_args(args), parser


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """

    args_bk = args
    args, parser = parse_args(args)
    if len(args_bk) == 0:
      parser.print_help()
      parser.exit()

    setup_logging(args.loglevel)
    _logger.info("Starting loon...")
    host = Host()

    # Deparse arguments
    if args.subparsers_name == 'add':
      _logger.info("Add command is detected.")
      if args.username is None or args.host is None:
        print("Error: username and host are both required in add command.")
        sys.exit(1)
      if args.name is None:
        args.name = args.username
      host.add(
        name=args.name,
        username=args.username,
        host=args.host,
        port=args.port)
      if args.switch_active:
        host.switch(
          name=args.name,
          username=args.username,
          host=args.host,
          port=args.port)
    elif args.subparsers_name == 'delete':
      _logger.info("Delete command is detected.")
      if args.username is None or args.host is None:
        if args.name is None:
          print("Error: either specify name or both username and host")
          sys.exit(1)
      host.delete(
        name=args.name,
        username=args.username,
        host=args.host,
        port=args.port)
    elif args.subparsers_name == 'switch':
      _logger.info("Switch command is detected.")
      if args.username is None or args.host is None:
        if args.name is None:
          print("Error: either specify name or both username and host")
          sys.exit(1)
      host.switch(
        name=args.name,
        username=args.username,
        host=args.host,
        port=args.port)
    elif args.subparsers_name == 'list':
      _logger.info("List command is detected.")
      host.list()
    elif args.subparsers_name == 'rename':
      _logger.info("Rename command is detected.")
      host.rename(args.old, args.new)
    elif args.subparsers_name == 'run':
      _logger.info("Run command is detected.")
      host.cmd(" ".join(args.commands))
    elif args.subparsers_name == 'upload':
      _logger.info("Upload command is detected.")
      host.connect(open_channel=False)
      host.upload(args.source, args.destination)

    _logger.info("loon ends here")


def run():
    """Entry point for console_scripts"""
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
import os

__version__ = "0.1.5"
__author__ = "ShixiangWang"
__copyright__ = "ShixiangWang"
__license__ = "mit"
# User can modify the following line
# to change the default path for host file
__host_file__ = "~/.config/loon/host.json"
# Get the absolute path for host file
# NEVER CHANGE IT!
__host_file__ = os.path.expanduser(__host_file__)
__privatekey_file__ = os.path.expanduser("~/.ssh/id_rsa")



# -*- coding: utf-8 -*-
"""Classes used in loon package"""

import sys
import os
import json
import socket
from datetime import datetime
from ssh2.session import Session
if __package__ == '' or __package__ is None:  # Use for test
    from __init__ import __host_file__
    from utils import create_parentdir, isfile, isdir, pretty_table
else:
    from loon import __host_file__
    from loon.utils import create_parentdir, isfile, isdir, pretty_table

class Host:
    """
    Representation of remote host
    """
    def __init__(self, hostfile = __host_file__):
        self.hostfile = hostfile
        self.load_hosts()
        return

    def load_hosts(self):
        """Load hosts from file"""
        if not isfile(self.hostfile):
            self.active_host = []
            self.available_hosts = []
        else:
            with open(self.hostfile, 'r') as f:
                hosts = json.load(f)
            self.active_host = hosts['active']
            self.available_hosts = hosts['available']

        if any(isinstance(i, list) for i in self.active_host):
            print("Error: more than one active host. Please check config file ~/.config/loon/host.json and modify or remove it if necessary.")

        # Python code to remove duplicate elements 
        def RemoveDups(duplicate): 
            final_list = [] 
            flag = False
            for num in duplicate: 
                if num not in final_list: 
                    final_list.append(num) 
                else: 
                    flag = True
            return final_list, flag
        
        self.available_hosts, flag = RemoveDups(self.available_hosts)
        
        if flag:
            # Save unique hosts immediately
            self.save_hosts()

        return

    def save_hosts(self):
        """Save hosts to file"""
        # if len(self.active_host)==0 or len(self.available_hosts)==0:
        #     raise ValueError("Cannot save to file due to null host.")
        hosts = {'active':self.active_host, 'available':self.available_hosts}
        if not isfile(self.hostfile):
            # Create parent dir if hostfile does not exist
            create_parentdir(self.hostfile)
        with open(self.hostfile, 'w') as f:
            json.dump(hosts, f)
        return

    def add(self, name, username, host, port=22):
        """Add a remote host"""
        info = [name, username, host, port]

        if info in self.available_hosts:
            print("=> Input host exists. Will not change.")
            return
        else:
            self.available_hosts.append(info)
            if len(self.active_host) == 0:
                self.active_host = info
            self.save_hosts()
            print("=> Added successfully!")
        return

    def host_check(self, name, username, host, port=22):
        host = []
        if name is not None:
            for h in self.available_hosts:
                if h[0] == name:
                    host = h.copy()
        else:
            info = [username, host, port]
            for h in self.available_hosts:
                if h[1:] == info:
                    host = h.copy()
        if len(host) == 0:
            print("=> Host does not exist, please check input with list command!")
            sys.exit(1)
        return host

    def delete(self, name, username, host, port=22):
        """Delete a remote host"""
        host2del = self.host_check(name, username, host, port)
        print("=> Removing host from available list...")
        self.available_hosts.remove(host2del)
        if host2del == self.active_host:
            print("=> Removing active host...")
            if len(self.available_hosts) > 0:
                self.active_host = self.available_hosts[0]
                print("=> Changing active host to %s" %self.active_host[0])
            else:
                self.active_host = []  # reset
                print("=> Reseting active host to []")
        self.save_hosts()
        return
    
    def switch(self, name, username, host, port=22):
        """Switch active host"""
        host2switch = self.host_check(name, username, host, port)
        self.active_host = host2switch
        self.save_hosts()
        print("=> Activated.")
        return

    def rename(self, old, new):
        """Rename host name"""
        host2rename = []        
        for index, h in enumerate(self.available_hosts):
            if h[0] == old:
                host2rename = h.copy()
                self.available_hosts[index][0] = new
        if len(host2rename) == 0:
            print("=> Host does not exist, please check input with list command!")
            sys.exit(1)
        if host2rename == self.active_host:
            self.active_host[0] = new
        self.save_hosts()
        return
        
        
    def list(self):
        """List all remote hosts"""

        title = ['Alias', 'Username', 'IP address', 'Port']
        content = self.available_hosts.copy()
        for host in content:
            if host == self.active_host:
                host[0] = '<'+host[0]+'>'
        pretty_table(title, content)
        print("<active host>")
        return
    
    def connect(self, privatekey_file="~/.ssh/id_rsa", passphrase='', open_channel=True):
        """Connect active host and open a session"""
        privatekey_file = os.path.expanduser(privatekey_file)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((self.active_host[2], self.active_host[3]))
        s = Session()
        s.handshake(sock)
        try:
            # Try using private key file first
            s.userauth_publickey_fromfile(self.active_host[1], privatekey_file, passphrase)
        except:
            # Use password to auth
            passwd = input('No private key found.\nEnter your password for %s: ' %self.active_host[1])
            s.userauth_password(self.active_host[1], passwd)
        self.session = s
        if open_channel:
            self.channel = self.session.open_session()
        return
    
    def cmd(self, commands):
        """Run command(s) in active remote host using channel session
        Therefore, `open_channel` in `connect` method must be `True` before using it.

        Args:
            commands ([str]): commands run on active remote host
        """
        self.connect()
        self.channel.execute(commands)

        size, errinfo = self.channel.read_stderr()
        if size > 0:
            print('An error is raised by remote host, please read the info:\n')
            print(errinfo.decode('utf-8'), end="")
            sys.exit(1)
        else:
            # Get output
            datalist = []
            size, data = self.channel.read()
            # Here data is byte type
            while size > 0:
                data = data.decode('utf-8')
                print(data, end='')
                datalist.append(data)
                size, data = self.channel.read()
        
        # Return a list containing output from commands 
        return datalist

    def upload(self, source, destination):
        """Upload files to active remote host
        Args:
            source [(list)]: list of files (directories) to remote host
            destination [(str)]: destination directory
        """
        print("Starting upload...")
        now = datetime.now()
        for i in source:
            info = os.stat(i)
            print("Uploading %s to %s" %(i, destination))
            destination = '/'.join([destination, os.path.basename(i)])
            chan = self.session.scp_send64(
                destination,
                info.st_mode & 0o777,
                info.st_size,
                info.st_mtime,
                info.st_atime
            )
            with open(i, 'rb') as local_fh:
                for data in local_fh:
                    chan.write(data)
        taken = datetime.now() - now
        print("Finished uploading in %s" %taken)
        return
     
        
    def download(self):
        """Download files to local machine from active remote host"""
        pass


class PBS:
    """
    Representation of PBS task
    """
    def __init__(self):
        pass
    def gen_template(self):
        pass
    def gen_pbs(self):
        pass
    def sub(self):
        """Submit all pbs tasks in a remote directory"""
        pass
    def check(self):
        """Check PBS task status"""
        pass

if __name__ == "__main__":
    host = Host()
    host.add(username="wsx", host="10.19.24.165")
    host.list()
    host.cmd("ls -l")
